#!/usr/bin/env python3

import freezerbox, pytest
from pathlib import Path
from freezerbox import load_config, ReagentConfig, MakerArgsConfig, Fields, cd
from more_itertools import one, first

from test_model import MockReagent

MOCK_CONFIG = Path(__file__).parent / 'mock_config'

class MockObj:
    pass

def test_config():
    with cd(MOCK_CONFIG):
        load_config.cache_clear()
        config = load_config()

        # Only check the values that are explicitly set by the test, because 
        # any other values could be affected by real configuration files 
        # present on the tester's system.

        assert config['use'] == 'db1'
        assert config['database']['db1']['type'] == 'type1'
        assert config['database']['db1']['option'] == 'option1'
        assert config['database']['db2']['type'] == 'type2'
        assert config['database']['db2']['option'] == 'option2'

    with cd(MOCK_CONFIG / 'subdir'):
        load_config.cache_clear()
        config = load_config()

        assert config['use'] == 'db2'
        assert config['database']['db1']['type'] == 'type1'
        assert config['database']['db1']['option'] == 'option1'
        assert config['database']['db2']['type'] == 'type2'
        assert config['database']['db2']['option'] == 'option2'

    load_config.cache_clear()


def test_reagent_config_tags_1():
    db = freezerbox.Database(name='a')
    db['x1'] = MockReagent(name='1')

    obj = MockObj()
    config = ReagentConfig()
    layer = one(config.load(obj))

    obj.db = db
    obj.tag = 'x1'

    assert layer.values['name'] == ['1']
    assert layer.location == 'a'

def test_reagent_config_tags_2():
    db = freezerbox.Database(name='a')
    db['x1'] = MockReagent(name='1')
    db['x2'] = MockReagent(name='2')

    obj = MockObj()
    config = ReagentConfig()
    layer = one(config.load(obj))

    obj.db = db
    obj.tag = 'x1', 'x2'

    assert layer.values['name'] == ['1', '2']
    assert layer.location == 'a'

def test_reagent_config_tags_not_found():
    db = freezerbox.Database(name='a')
    db['x1'] = MockReagent(name='1')

    obj = MockObj()
    config = ReagentConfig()
    layer = one(config.load(obj))

    obj.db = db
    obj.tag = 'x2'

    with pytest.raises(KeyError):
        layer.values['name']

    assert layer.location == 'a'

def test_reagent_config_tags_not_parseable():
    db = freezerbox.Database(name='a')
    db['x1'] = MockReagent(name='1')

    obj = MockObj()
    config = ReagentConfig()
    layer = one(config.load(obj))

    obj.db = db
    obj.tag = 'not-a-tag'

    with pytest.raises(KeyError):
        layer.values['name']

    assert layer.location == 'a'

def test_reagent_config_key_not_found():
    db = freezerbox.Database(name='a')
    db['x1'] = MockReagent(name='1')

    obj = MockObj()
    config = ReagentConfig()
    layer = one(config.load(obj))

    obj.db = db
    obj.tag = 'x1'

    with pytest.raises(KeyError):
        layer.values['not-a-key']

    assert layer.location == 'a'

def test_reagent_config_transform():
    db = freezerbox.Database(name='a')
    db['x1'] = MockReagent(name='1')
    db['x2'] = MockReagent(name='2')

    obj = MockObj()
    config = ReagentConfig(transform=first)
    layer = one(config.load(obj))

    obj.db = db
    obj.tag = ['x1', 'x2']

    assert layer.values['name'] == '1'
    assert layer.location == 'a'

def test_reagent_config_db_autoload(monkeypatch):
    db = freezerbox.Database(name='a')
    db['x1'] = MockReagent(name='1')
    monkeypatch.setattr(freezerbox.model, 'load_db', lambda: db)

    obj = MockObj()
    config = ReagentConfig()
    layer = one(config.load(obj))

    obj.tag = 'x1'

    assert layer.values['name'] == ['1']
    assert layer.location == 'a'

def test_reagent_config_db_not_found():
    obj = MockObj()
    config = ReagentConfig(autoload_db=False)
    layer = one(config.load(obj))

    obj.tag = 'x1'

    with pytest.raises(KeyError, match="no freezerbox database found"):
        layer.values['name']

    assert layer.location == '*no database loaded*'

def test_reagent_config_empty_db():
    db = freezerbox.Database(name='a')

    obj = MockObj()
    config = ReagentConfig()
    layer = one(config.load(obj))

    obj.db = db
    obj.tag = []

    assert layer.values['x'] == []
    assert layer.values.db is db
    assert layer.location == 'a'

def reagent_config_from_ctor():
    return ReagentConfig(
            db_getter=lambda self: self.my_db,
            tag_getter=lambda self: self.my_tag,
    )

def reagent_config_from_subclass():

    class MyConfig(ReagentConfig):
        db_getter = lambda self: self.my_db
        tag_getter = lambda self: self.my_tag

    return MyConfig()

@pytest.mark.parametrize(
        'config_factory', [
            reagent_config_from_ctor,
            reagent_config_from_subclass,
        ]
)
def test_reagent_config_getters(config_factory):
    db = freezerbox.Database(name='a')
    db['x1'] = MockReagent(name='1')

    obj = MockObj()
    config = config_factory()
    layer = one(config.load(obj))

    obj.my_db = db
    obj.my_tag = 'x1'

    assert layer.values['name'] == ['1']
    assert layer.location == 'a'


def test_maker_args_config_synthesis():
    db = freezerbox.Database(name='loc')
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['a'], {'b': 'c'}),
    )
    i1 = x1.make_intermediate(0)

    obj = MockObj()
    config = MakerArgsConfig()
    layer = one(config.load(obj))

    obj.products = [i1]

    assert layer.values[0] == 'a'
    assert layer.values['b'] == 'c'
    assert layer.values.product is i1
    assert layer.location == 'loc'

def test_maker_args_config_cleanup():
    db = freezerbox.Database(name='loc')
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['a'], {'b': 'c'}),
            cleanups=[Fields(['d'], {'e': 'f'})],
    )
    i1 = x1.make_intermediate(1)

    obj = MockObj()
    config = MakerArgsConfig()
    layer = one(config.load(obj))

    obj.products = [i1]

    assert layer.values[0] == 'd'
    assert layer.values['e'] == 'f'
    assert layer.values.product is i1
    assert layer.values.precursor is i1.precursor
    assert layer.location == 'loc'

def maker_args_config_from_ctor():
    return MakerArgsConfig(
            products_getter=lambda self: self.my_products,
    )

def maker_args_config_from_subclass():

    class MyConfig(MakerArgsConfig):
        products_getter = lambda self: self.my_products

    return MyConfig()

@pytest.mark.parametrize(
        'config_factory', [
            maker_args_config_from_ctor,
            maker_args_config_from_subclass,
        ]
)
def test_maker_args_config_getters_inherit(config_factory):
    db = freezerbox.Database(name='loc')
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['a'], {'b': 'c'}),
            cleanups=[Fields(['d'], {'e': 'f'})],
    )
    i1 = x1.make_intermediate(0)

    obj = MockObj()
    config = config_factory()
    layer = one(config.load(obj))

    obj.my_products = [i1]

    assert layer.values[0] == 'a'
    assert layer.values['b'] == 'c'
    assert layer.values.product is i1
    assert layer.location == 'loc'

def test_maker_args_config_err():
    db = freezerbox.Database(name='loc')
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['a'], {'b': 'c'}),
    )
    i1 = x1.make_intermediate(0)

    obj = MockObj()
    config = MakerArgsConfig()
    layer = one(config.load(obj))

    obj.products = []

    with pytest.raises(KeyError, match="expected 1 product, found 0"):
        layer.values[0]




#!/usr/bin/env python3

import pytest
import parametrize_from_file

from freezerbox import Tag
from freezerbox.model import *
from freezerbox.utils import *
from stepwise import Quantity
from schema_helpers import *

class PrefixParams(Params):
    args = 'args, expected'
    params = [
            # 'x' is from the various mock reagents.
            ([],                                    'xbrfpo'),
            ([Reagent],                             'xbrfpo'),
            ([Buffer],                              'b'),
            ([Protein],                             'r'),
            ([Protein, NucleicAcid],                'xrfpo'),
            ([NucleicAcid],                         'xfpo'),
            ([NucleicAcid, Plasmid],                'xfpo'),
            ([NucleicAcid, Oligo],                  'xfpo'),
            ([NucleicAcid, Plasmid, Oligo],         'xfpo'),
            ([Plasmid],                             'p'),
            ([Oligo],                               'o'),
            ([Plasmid, Oligo],                      'po'),
    ]

@parametrize_from_file
def test_normalize_seq(raw_seq, expected):
    assert normalize_seq(raw_seq) == expected

@PrefixParams.parametrize
def test_get_tag_prefixes(args, expected):
    assert get_tag_prefixes(*args) == set(expected)

@PrefixParams.parametrize
def test_get_tag_pattern(args, expected):
    assert re.fullmatch(
            fr'\[[{expected}]+\]\\d\+',
            get_tag_pattern(*args),
    )

@parametrize_from_file(
        schema=Schema({
            'tag_str': str,
            **error_or(**{
                'expected': {
                    'type': str,
                    'id': Coerce(int),
                },
            }),
        }),
)
def test_parse_tag(tag_str, expected, error):
    with error:
        assert parse_tag(tag_str) == Tag(**expected)

@parametrize_from_file(
        schema=Schema({
            'bool_str': str,
            **error_or(**{
                'expected': eval,
            }),
        }),
)
def test_parse_bool(bool_str, expected, error):
    with error:
        assert parse_bool(bool_str) == expected

@parametrize_from_file(
        schema=Schema({
            'time_str': str,
            **error_or(**{
                'expected': Coerce(float),
            }),
        }),
)
def test_parse_time_s(time_str, expected, error):
    with error:
        assert parse_time_s(time_str) == pytest.approx(expected)

@parametrize_from_file(
        schema=Schema({
            'temp_str': str,
            **error_or(**{
                'expected': Coerce(float),
            }),
        }),
)
def test_parse_temp_C(temp_str, expected, error):
    with error:
        assert parse_temp_C(temp_str) == pytest.approx(expected)

@parametrize_from_file(
        schema=Schema({
            'vol_str': str,
            **error_or(**{
                'expected': Coerce(float),
            }),
        }),
)
def test_parse_volume_uL(vol_str, expected, error):
    with error:
        assert parse_volume_uL(vol_str) == expected

@parametrize_from_file(
        schema=Schema({
            'mass_str': str,
            **error_or(**{
                'expected': Coerce(float),
            }),
        }),
)
def test_parse_mass_ug(mass_str, expected, error):
    with error:
        assert parse_mass_ug(mass_str) == expected

@parametrize_from_file(
        schema=Schema({
            'conc_str': str,
            'mw': Coerce(float),
            'expected_nM': Coerce(float),
            'expected_ng_uL': Coerce(float),
        }),
)
def test_parse_conc(conc_str, mw, expected_nM, expected_ng_uL):
    from itertools import combinations

    conc = {
            'nM': parse_conc_nM(conc_str, mw),
            'uM': parse_conc_uM(conc_str, mw),
            'ng/uL': parse_conc_ng_uL(conc_str, mw),
    }
    expected = {
            'nM': expected_nM,
            'uM': expected_nM / 1000,
            'ng/uL': expected_ng_uL,
    }

    for k in conc:
        assert conc[k] == pytest.approx(expected[k])

    for k1, k2 in combinations(conc, 2):
        q_given = Quantity(conc[k1], k1)
        q_expected = Quantity(expected[k2], k2)
        q_converted = convert_conc_unit(q_given, mw, k2)
        assert q_converted == pytest.approx(
                q_expected,
                abs=Quantity(1e-6, k2),
        )

@parametrize_from_file(
        schema=Schema({
            'conc_str': str,
            'mw': Coerce(float),
            'error': error,
        }),
)
def test_parse_conc_err(conc_str, mw, error):
    with error:
        parse_conc_nM(conc_str, mw)
    with error:
        parse_conc_uM(conc_str, mw)
    with error:
        parse_conc_ng_uL(conc_str, mw)

@parametrize_from_file(
        schema=Schema({
            'size_str': str,
            **error_or(**{
                'expected': Coerce(int),
            }),
        }),
)
def test_parse_size_bp(size_str, expected, error):
    with error:
        assert parse_size_bp(size_str) == expected

@parametrize_from_file(
        schema=Schema({
            'items': eval,
            **error_or(**{
                'expected': eval,
            }),
        }),
)
def test_unanimous(items, expected, error):
    with error:
        assert unanimous(items) == expected

@parametrize_from_file(schema=Schema({str: eval}))
def test_join_lists(given, expected):
    assert join_lists(given) == expected

@parametrize_from_file(schema=Schema({str: eval}))
def test_join_dicts(given, expected):
    assert join_dicts(given) == expected

@parametrize_from_file(schema=Schema({str: eval}))
def test_join_sets(given, expected):
    assert join_sets(given) == expected




#!/usr/bin/env python3

import freezerbox
import parametrize_from_file

from freezerbox import Database, parse_fields
from freezerbox.stepwise.make import Make
from schema_helpers import *
from mock_model import *
from os import getcwd

@parametrize_from_file(
        schema=Schema({
            'db': eval_db,
            'expected': empty_ok([str]),
        }),
)
def test_make(db, expected, disable_capture, mock_plugins):
    cwd = getcwd()

    tags = list(db.keys())
    app = Make(db, tags)

    with disable_capture:
        assert app.protocol.steps == expected

    assert getcwd() == cwd

@parametrize_from_file(
        schema=Schema({
            'maker': str,
            'expected': {str: eval_freezerbox},
        }),
)
def test_builtin_maker_attrs(maker, expected, disable_capture):
    db = Database()
    db['x1'] = x1 = MockReagent(
            synthesis=parse_fields(maker),
    )

    with disable_capture:
        for attr, value in expected.items():
            assert getattr(x1.synthesis_maker, attr) == value

@pytest.fixture
def disable_capture(pytestconfig):
    # Equivalent to `pytest -s`, but temporary.
    # This is necessary because even `capfd.disabled()` leaves stdin in a state 
    # that somehow interferes with the redirection we're trying to do.

    class suspend_guard:

        def __init__(self):
            self.capmanager = pytestconfig.pluginmanager.getplugin('capturemanager')

        def __enter__(self):
            self.capmanager.suspend_global_capture(in_=True)

        def __exit__(self, _1, _2, _3):
            self.capmanager.resume_global_capture()

    yield suspend_guard()




#!/usr/bin/env python3

import re, os
from stepwise import Quantity
from more_itertools import split_when
from contextlib import contextmanager
from .errors import ParseError, only_raise

no_default = object()

def normalize_seq(raw_seq):
    # Ignore nonstandard nucleotides; they're too hard to deal with properly.

    # This regular expression works because `re.sub()` only substitutes the 
    # left-most occurrence of any overlapping patterns.  The non-greedy * is 
    # necessary to avoid eliding everything between the first and last 
    # nonstandard nucleotide.
    seq = re.sub(r'/.*?/', 'X', raw_seq)

    return seq.upper()

def get_tag_prefixes(*args):
    from .model import Reagent

    def add(prefixes, cls):
        if cls.tag_prefix:
            prefixes.add(cls.tag_prefix)
        for subcls in cls.__subclasses__():
            add(prefixes, subcls)

    prefixes = set()
    for cls in args or [Reagent]:
        add(prefixes, cls)
    
    return prefixes

def get_tag_pattern(*args):
    prefix_chars = ''.join(sorted(get_tag_prefixes(*args)))
    return fr'[{prefix_chars}]\d+'

@only_raise(ParseError)
def parse_tag(tag_str):
    from .model import Tag, get_tag_prefixes

    if isinstance(tag_str, Tag):
        return tag_str
    if isinstance(tag_str, tuple):
        return Tag(*tag_str)

    pfo = get_tag_prefixes()
    if m := re.fullmatch(fr'\s*(?P<type>[{pfo}])(?P<id>\d+)\s*', tag_str):
        return Tag(m.group('type'), int(m.group('id')))
    else:
        raise ParseError(f"expected a tag (e.g. 'p100'), not {tag_str!r}")

@only_raise(ParseError)
def parse_bool(bool_str):
    if bool_str.lower() in ('1', 'y', 'yes', 'true'):
        return True
    if bool_str.lower() in ('0', 'n', 'no', 'false'):
        return False

    raise ParseError(f"can't interpret {bool_str!r} as a bool")

@only_raise(ParseError)
def parse_time_s(time_str):
    time_units = {
            's':        1,
            'sec':      1,
            'second':   1,
            'seconds':  1,
            'm':        60,
            'min':      60,
            'minute':   60,
            'minutes':  60,
            'h':        60*60,
            'hr':       60*60,
            'hour':     60*60,
            'hours':    60*60,
    }
    time_pattern_1 = fr'(?P<time>\d+)\s*(?P<unit>{"|".join(time_units)})'
    time_pattern_2 = fr'(?P<min>\d+)m(?P<sec>\d+)'

    if m := re.fullmatch(time_pattern_1, time_str):
        return int(m.group('time')) * time_units[m.group('unit')]
    if m := re.fullmatch(time_pattern_2, time_str):
        return 60 * int(m.group('min')) + int(m.group('sec'))

    raise ParseError(f"can't interpret {time_str!r} as a time, did you forget a unit?")

@only_raise(ParseError)
def parse_time_m(time_str):
    return parse_time_s(time_str) / 60

@only_raise(ParseError)
def parse_time_h(time_str):
    return parse_time_s(time_str) / 3600

@only_raise(ParseError)
def parse_temp_C(temp_str):
    temp_pattern = fr'(?P<temp>[0-9.]+)\s*°?C'

    if m := re.fullmatch(temp_pattern, temp_str):
        return float(m.group('temp'))

    raise ParseError(f"can't interpret {temp_str!r} as a temperature, did you forget a unit?")

@only_raise(ParseError)
def parse_volume_uL(vol_str):
    vol_pattern = fr'(?P<vol>\d+)\s*(?P<si_prefix>[nµum])L'
    si_prefixes = {
            'n': 1e-3,
            'u': 1,
            'µ': 1,
            'm': 1e3,
    }

    if m := re.fullmatch(vol_pattern, vol_str):
        return float(m.group('vol')) * si_prefixes[m.group('si_prefix')]

    raise ParseError(f"can't interpret {vol_str!r} as a volume, did you forget a unit?")

@only_raise(ParseError)
def parse_mass_ug(mass_str):
    mass_pattern = fr'(?P<mass>\d+)\s*(?P<si_prefix>[nµum])g'
    si_prefixes = {
            'n': 1e-3,
            'u': 1,
            'µ': 1,
            'm': 1e3,
    }

    if m := re.fullmatch(mass_pattern, mass_str):
        return float(m.group('mass')) * si_prefixes[m.group('si_prefix')]

    raise ParseError(f"can't interpret {mass_str!r} as a mass, did you forget a unit?")

@only_raise(ParseError)
def parse_conc_nM(conc_str, mw):
    conc_pattern = r'(?P<conc>\d+)\s?(?P<unit>[nuµ]M|ng/[uµ]L)'
    unit_conversion = {
            'nM': 1,
            'uM': 1e3,
            'µM': 1e3,
            'ng/uL': 1e6 / mw,
            'ng/µL': 1e6 / mw,
    }

    if m := re.match(conc_pattern, conc_str):
        return float(m.group('conc')) * unit_conversion[m.group('unit')]
    else:
        raise ParseError(f"can't interpret {conc_str!r} as a concentration, did you forget a unit?")

@only_raise(ParseError)
def parse_conc_uM(conc_str, mw):
    return parse_conc_nM(conc_str, mw) / 1000

@only_raise(ParseError)
def parse_conc_ng_uL(conc_str, mw):
    conc_pattern = r'(?P<conc>\d+)\s?(?P<unit>[nuµ]M|ng/[uµ]L)'
    unit_conversion = {
            'ng/uL': 1,
            'ng/µL': 1,
            'nM': mw / 1e6,
            'uM': mw / 1e3,
            'µM': mw / 1e3,
    }

    if m := re.match(conc_pattern, conc_str):
        return float(m.group('conc')) * unit_conversion[m.group('unit')]
    else:
        raise ParseError(f"can't interpret {conc_str!r} as a concentration, did you forget a unit?")

@only_raise(ParseError)
def convert_conc_unit(conc, mw, new_unit):
    molar_conversion_factors = {
            'pM': 1e12,
            'nM': 1e9,
            'uM': 1e6,
            'µM': 1e6,
            'mM': 1e3,
            'M': 1,
    }
    mass_volume_conversion_factors = {
            'ng/uL': 1e3,
            'ng/µL': 1e3,
            'µg/µL': 1,
            'ug/uL': 1,
            'mg/mL': 1,
    }

    def pick_conversion_factors():
        if mw is not None:
            return {
                    **molar_conversion_factors,
                    **{
                        k: mw * v
                        for k, v in mass_volume_conversion_factors.items()
                    },
            }

        if conc.unit in molar_conversion_factors:
            return molar_conversion_factors

        if conc.unit in mass_volume_conversion_factors:
            return mass_volume_conversion_factors

        return {}

    return conc.convert_unit(new_unit, pick_conversion_factors())

@only_raise(ParseError)
def parse_size_bp(size_str):
    bp_parsers = [
            (
                fr'(?P<size>\d+)\s*bp',
                int,
            ), (
                fr'(?P<size>\d+.?\d*)\s*kb',
                lambda x: int(float(x) * 1000),
            ),
    ]

    for pattern, size_from_str in bp_parsers:
        if m := re.fullmatch(pattern, size_str):
            return size_from_str(m.group('size'))

    raise ParseError(f"can't interpret {size_str!r} as a size in base pairs, did you forget a unit?")

def unanimous(items):
    it = iter(items)

    try:
        value = next(it)
    except StopIteration:
        raise ValueError("empty iterable")

    for next_value in it:
        if next_value != value:
            raise ValueError(f"found multiple values: {value!r}, {next_value!r}")

    return value

def join_lists(x):
    from itertools import chain
    return list(chain(*x))

def join_dicts(x):
    from collections import ChainMap
    return dict(ChainMap(*x))

def join_sets(x):
    return set.union(set(), *x)

@contextmanager
def cd(dir):
    try:
        prev_cwd = os.getcwd()
        os.chdir(dir)
        yield

    finally:
        os.chdir(prev_cwd)




#!/usr/bin/env python3

import appdirs
import appcli

from pathlib import Path
from configurator import Config
from functools import lru_cache
from more_itertools import one, always_iterable

from .utils import parse_tag
from .errors import QueryError, ParseError, only_raise

@lru_cache
def load_config():
    config = Config()
    cwd = Path.cwd()
    dirs = [cwd, *cwd.parents]

    for dir in reversed(dirs):
        paths = [
                dir / '.config' / 'freezerbox' / 'conf.toml',
                dir / '.freezerboxrc',
        ]
        for path in paths:
            if path.exists():
                subconf = Config.from_path(path, parser='toml')
                config.merge(subconf)

        dir = dir.parent

    return config.data

class ReagentConfig:
    autoload = True
    autoload_db = True
    db_getter = lambda obj: obj.db
    tag_getter = lambda obj: obj.tag
    transform = list

    class QueryHelper:

        def __init__(self, config, obj):
            self.config = config
            self.obj = obj
            self.db = None

        def __getitem__(self, key):
            # First: See if the object has a database attribute.  If it does, 
            # it costs nothing to access it, and it will allow us to include 
            # the path to the database in error messages if any subsequent 
            # steps fail.

            if self.db is None:
                try:
                    self.db = self.config.db_getter(self.obj)
                except AttributeError:
                    pass

            # Second: Parse the tags before loading the database.  Loading the 
            # database is expensive, and if the tags won't be in the database 
            # anyways, there's no reason to waste the time.
            
            try:
                tags = self.config.tag_getter(self.obj)
            except AttributeError as err:
                raise KeyError from err

            try:
                tags = [
                        parse_tag(x)
                        for x in always_iterable(tags)
                ]
            except ParseError as err:
                raise KeyError from err

            # Third: Load the database.

            if self.db is None and self.config.autoload_db:
                from .model import load_db
                self.db = load_db()

            if self.db is None:
                raise KeyError("no freezerbox database found")

            # Fourth: Lookup the key as an attribute of the selected reagents.

            try:
                values = [
                        getattr(self.db[x], key)
                        for x in tags
                ]

            except (QueryError, AttributeError) as err:
                raise KeyError from err

            return self.config.transform(values)

        def get_location(self):
            return self.db.name if self.db is not None else "*no database loaded*"

    def __init__(self, tag_getter=None, db_getter=None, autoload_db=None, transform=None):
        cls = self.__class__

        # Access the getter/transform functions through the class.  If accessed 
        # via the instance, they would become bound and would require a self 
        # argument. 

        self.db_getter = db_getter or cls.db_getter
        self.tag_getter = tag_getter or cls.tag_getter
        self.autoload_db = autoload_db if autoload_db is not None else self.autoload_db
        self.transform = transform or cls.transform

    def load(self, obj):
        helper = self.QueryHelper(self, obj)
        yield appcli.Layer(
                values=helper,
                location=helper.get_location,
        )


class MakerArgsConfig:
    autoload = False
    products_getter = lambda obj: obj.products

    class QueryHelper:

        def __init__(self, config, obj):
            self.config = config
            self.obj = obj

        @only_raise(KeyError)
        def __getitem__(self, key):
            return self.product.maker_args[key]

        @property
        def product(self):
            products = self.config.products_getter(self.obj)
            try:
                return one(products)
            except ValueError:
                err = QueryError(
                        lambda e: f"expected 1 product, found {len(products)}",
                        products=products,
                )
                raise err from None

        @property
        def precursor(self):
            return self.product.precursor

        def get_location(self):
            return self.product.db.name


    def __init__(self, db_getter=None, products_getter=None):
        cls = self.__class__

        # Access the getter through the class.  If accessed via the instance, 
        # it would become bound and would require a self argument. 

        self.products_getter = products_getter or cls.products_getter

    def load(self, obj):
        helper = self.QueryHelper(self, obj)
        yield appcli.Layer(
                values=helper,
                location=helper.get_location,
        )

